from abc import ABC
from datetime import datetime
from typing import Any

from jwt.exceptions import PyJWTError
from pydantic import BaseModel, EmailStr
from ravyn.conf import settings
from ravyn.exceptions import AuthenticationError, NotAuthorized

from ravyn_simple_jwt.schemas import AccessToken, RefreshToken
from ravyn_simple_jwt.token import Token


class BaseBackendAuthentication(ABC, BaseModel):
    """
    Base for all authentication backends.
    """

    async def authenticate(self) -> dict[str, str] | Any:
        raise NotImplementedError("All backends must implement the `authenticate()` method.")


class BaseRefreshAuthentication(ABC, BaseModel):
    """
    Base for all refresh backends.
    """

    async def refresh(self) -> dict[str, str] | Any:
        raise NotImplementedError("All refresh backends must implement the `refresh()` method.")


class BackendEmailAuthentication(BaseBackendAuthentication):
    """
    Utility class that helps with the authentication process using email and password.
    """

    email: EmailStr
    password: str

    async def authenticate(self) -> dict[str, str] | Any: ...


class BackendUsernameAuthentication(BaseBackendAuthentication):
    """
    Utility class that helps with the authentication process using username and password.
    """

    username: str
    password: str

    async def authenticate(self) -> dict[str, str] | Any: ...


class RefreshAuthentication(BaseRefreshAuthentication):
    """
    Refreshes the access token given a refresh token of a given user.

    This object does not perform any DB action, instead, uses the existing refresh
    token to generate a new access.
    """

    token: RefreshToken

    async def refresh(self) -> AccessToken:
        token = self.token.refresh_token

        try:
            token = Token.decode(
                token=token,
                key=settings.simple_jwt.signing_key,
                algorithms=[settings.simple_jwt.algorithm],
            )  # type: ignore
        except PyJWTError as e:
            raise AuthenticationError(str(e)) from e

        if token.token_type != settings.simple_jwt.refresh_token_name:
            raise NotAuthorized(detail="Only refresh tokens are allowed.")

        # Apply the maximum living time
        expiry_date = datetime.now() + settings.simple_jwt.access_token_lifetime

        # New token object
        new_token = Token(sub=token.sub, exp=expiry_date)

        claims_extra = {"token_type": settings.simple_jwt.access_token_name}

        # Encode the token
        access_token = new_token.encode(
            key=settings.simple_jwt.signing_key,
            algorithm=settings.simple_jwt.algorithm,
            claims_extra=claims_extra,
        )

        return AccessToken(access_token=access_token)



from ravyn import Gateway

from ravyn_simple_jwt.views import refresh_token, signin

route_patterns = [
    Gateway(handler=signin, name="simplejwt-signin"),
    Gateway(handler=refresh_token, name="simplejwt-refresh"),
]



from ravyn import JSONResponse, post, status
from ravyn.conf import settings
from ravyn.openapi.datastructures import OpenAPIResponse


@post(
    path=settings.simple_jwt.signin_url,
    summary=settings.simple_jwt.signin_summary,
    description=settings.simple_jwt.signin_description,
    status_code=status.HTTP_200_OK,
    security=settings.simple_jwt.security,
    tags=settings.simple_jwt.tags,
    responses={200: OpenAPIResponse(model=settings.simple_jwt.token_model)},
)
async def signin(data: settings.simple_jwt.login_model) -> JSONResponse:  # type: ignore
    """
    Login a user and returns a JWT token, else raises ValueError.
    """
    auth = settings.simple_jwt.backend_authentication(**data.model_dump())
    access_tokens: dict[str, str] = await auth.authenticate()
    return JSONResponse(access_tokens)


@post(
    path=settings.simple_jwt.refresh_url,
    summary=settings.simple_jwt.refresh_summary,
    description=settings.simple_jwt.refresh_description,
    security=settings.simple_jwt.security,
    tags=settings.simple_jwt.tags,
    status_code=status.HTTP_200_OK,
    responses={200: OpenAPIResponse(model=settings.simple_jwt.access_token_model)},
)
async def refresh_token(
    payload: settings.simple_jwt.refresh_model,  # type: ignore
) -> settings.simple_jwt.access_token_model:  # type: ignore
    authentication = settings.simple_jwt.backend_refresh(token=payload)
    access_token = await authentication.refresh()
    return access_token



from collections.abc import Sequence
from typing import TYPE_CHECKING, Annotated, Any, Optional

from ravyn import ChildRavyn, Ravyn
from ravyn.core.interceptors.types import Interceptor
from ravyn.permissions.types import Permission
from ravyn.pluggables import Extension
from ravyn.routing.router import Include
from ravyn.types import Dependencies, ExceptionHandlerMap, Middleware
from typing_extensions import Doc

if TYPE_CHECKING:
    from ravyn.types import SettingsType


class SimpleJWTExtension(Extension):
    """
    The pluggable version of Ravyn Simple JWT.

    This Pluggable can and should be used if you want to add
    the package independently as a ChilRavyn.

    **Example**

    ```python
    from ravyn import Ravyn, Pluggable
    from ravyn_simple_jwt.extension import SimpleJWTExtension


    app = Ravyn(
        pluggables={
            "simple-jwt": Pluggable(SimpleJWTExtension, path="/auth"),
        },
    )
    ```
    """

    def __init__(self, app: Optional["Ravyn"] = None, **kwargs: Any):
        super().__init__(app, **kwargs)
        self.app = app
        self.kwargs = kwargs

    def extend(
        self,
        path: Annotated[
            str | None,
            Doc(
                """
                Relative path of the Plugable.
                The path can contain parameters in a dictionary like format
                and if the path is not provided, it will default to `/`.

                **Example**

                ```python
                Pluugable(SimpleJWTExtension, path="/{age: int}"))
                ```
                """
            ),
        ] = "/simple-jwt",
        name: Annotated[
            str | None,
            Doc(
                """
                The name for the Gateway. The name can be reversed by `url_path_for()`.
                """
            ),
        ] = None,
        settings_module: Annotated[
            Optional["SettingsType"],
            Doc(
                """
                Alternative settings parameter. This parameter is an alternative to
                `RAVYN_SETTINGS_MODULE` way of loading your settings into an Ravyn application.

                When the `settings_module` is provided, it will make sure it takes priority over
                any other settings provided for the instance.

                Read more about the [settings module](https://ravyn.dev/application/settings/)
                and how you can leverage it in your application.

                !!! Tip
                    The settings module can be very useful if you want to have, for example, a
                    [ChildRavyn](https://ravyn.dev/routing/router/?h=childe#child-ravyn-application) that needs completely different settings
                    from the main app.

                    Example: A `ChildRavyn` that takes care of the authentication into a cloud
                    provider such as AWS and handles the `boto3` module.
                """
            ),
        ] = None,
        middleware: Annotated[
            Sequence["Middleware"] | None,
            Doc(
                """
                A list of middleware to run for every request. The middlewares of a Gateway will be checked from top-down or [Starlette Middleware](https://www.starlette.io/middleware/) as they are both converted internally. Read more about [Python Protocols](https://peps.python.org/pep-0544/).
                """
            ),
        ] = None,
        dependencies: Annotated[
            Optional["Dependencies"],
            Doc(
                """
                A dictionary of string and [Inject](https://ravyn.dev/dependencies/) instances enable application level dependency injection.
                """
            ),
        ] = None,
        exception_handlers: Annotated[
            Optional["ExceptionHandlerMap"],
            Doc(
                """
                A dictionary of [exception types](https://ravyn.dev/exceptions/) (or custom exceptions) and the handler functions on an application top level. Exception handler callables should be of the form of `handler(request, exc) -> response` and may be be either standard functions, or async functions.
                """
            ),
        ] = None,
        interceptors: Annotated[
            list["Interceptor"] | None,
            Doc(
                """
                A list of [interceptors](https://ravyn.dev/interceptors/) to serve the application incoming requests (HTTP and Websockets).
                """
            ),
        ] = None,
        permissions: Annotated[
            list["Permission"] | None,
            Doc(
                """
                A list of [permissions](https://ravyn.dev/permissions/) to serve the application incoming requests (HTTP and Websockets).
                """
            ),
        ] = None,
        include_in_schema: Annotated[
            bool | None,
            Doc(
                """
                Boolean flag indicating if it should be added to the OpenAPI docs.
                """
            ),
        ] = True,
        enable_openapi: Annotated[
            bool | None,
            Doc(
                """
                Boolean flag indicating if the OpenAPI documentation should
                be generated or not.

                When `False`, no OpenAPI documentation is accessible.

                !!! Tip
                    Disable this option if you run in production and no one should access the
                    documentation unless behind an authentication.
                ```
                """
            ),
        ] = True,
    ) -> None:
        """
        The extend() default from the Pluggable interface allowing to pass extra parameters
        to the initialisation.

        **Example**

        ```python
        from ravyn import Ravyn, Pluggable
        from ravyn_simple_jwt.extension import SimpleJWTExtension


        app = Ravyn(
            pluggables={
                "simple-jwt": Pluggable(
                    SimpleJWTExtension,
                    path="/auth",
                    settings_module=...,
                    middleware=...,
                    permissions=...,
                    interceptors=...,
                ),
            },
        )
        ```
        """
        simple_jwt = ChildRavyn(
            routes=[
                Include(namespace="ravyn_simple_jwt.urls"),
            ],
            middleware=middleware,
            dependencies=dependencies,
            exception_handlers=exception_handlers,
            interceptors=interceptors,
            permissions=permissions,
            include_in_schema=include_in_schema,
            enable_openapi=enable_openapi,
            settings_module=settings_module,
        )
        self.app.add_child_ravyn(
            path=path,
            child=simple_jwt,
            name=name,
            include_in_schema=include_in_schema,
        )



from pydantic import BaseModel, EmailStr


class AccessToken(BaseModel):
    """
    The representation of an access token.

    When `model_dump()` is called, it will generate a python like
    dictionary.

    ```python
    {"access_token": ...}
    ```
    """

    access_token: str


class RefreshToken(BaseModel):
    """
    The representation of a refresh token.

    When `model_dump()` is called, it will generate a python like
    dictionary.

    ```python
    {"refresh_token": ...}
    ```
    """

    refresh_token: str


class TokenAccess(AccessToken, RefreshToken):
    """
    The representation of token access used by the signin response.

    When `model_dump()` is called, it will generate a python like
    dictionary.

    ```python
    {
        "access_token": ...,
        "refresh_token": ...,
    }
    ```
    """

    ...


class LoginEmailIn(BaseModel):
    """
    The representation of a login payload used by the signin endpoint when using an
    email backend for validation.
    When the endpoint is called via `POST` it should contain the following.


    ```python
    {"email": ..., "password": ...}
    ```
    """

    email: EmailStr
    password: str


class LoginUserIn(BaseModel):
    """
    The representation of a login payload used by the signin endpoint when using an username backend for validation.
    When the endpoint is called via `POST` it should contain the following.


    ```python
    {"username": ..., "password": ...}
    ```
    """

    username: str
    password: str



from typing import Annotated

from ravyn.security.jwt.token import Token as RavynToken
from typing_extensions import Doc


class Token(RavynToken):
    """
    Token implementation with an extra field
    `token_type`. This attribute will allow
    the distintion of the type of token being generated.

    This token is an extension of the `ravyn.security.jwt.Token` object
    and adds some extras that are used by the default Refresh backend provided
    by the package.

    !!! Note
        You are not entitled to use this object at all in your backends but if you
        are to use the examples given and the defaults without wasting too much time
        the package examples use this object to classiify the `token_type` in the claims.
    """

    token_type: Annotated[
        str | None,
        Doc(
            """
            A string value classifying the type of token being generated and used
            for the claims.

            It can be something like `access_token` or `access` or
            `refresh_token` or `refresh` or any other string value
            that can help you distinguish the token being generated in the
            claims when `decode()` and `encode()` are called.
            """
        ),
    ] = None



from typing import Annotated, Any

from pydantic import BaseModel
from ravyn.core.config.jwt import JWTConfig
from ravyn.security.http import HTTPBearer
from typing_extensions import Doc

from ravyn_simple_jwt.backends import BaseBackendAuthentication, BaseRefreshAuthentication
from ravyn_simple_jwt.schemas import AccessToken, LoginEmailIn, RefreshToken, TokenAccess

security = HTTPBearer()


class SimpleJWT(JWTConfig):
    """
    A subclass of [JWTConfig](https://ravyn.dev/configurations/jwt/).

    When uwing this object, default values can be overridden to any application
    preference.

    **Example**

    ```python
    from ravyn import Ravyn, RavynAPISettings
    from ravyn_simple_jwt.config import SimpleJWT


    class AppSettings(RavynAPISettings):
        simple_jwt: SimpleJWT = SimpleJWT(
            signing_key=settings.secret_key,
            backend_authentication=...,
            backend_refresh=...,
        )
    ```
    """

    backend_authentication: Annotated[
        type[BaseBackendAuthentication],
        Doc(
            """
            The backend authentication being used by the system. A subclass of `ravyn_simple_jwt.backends.BaseBackendAuthentication`.

            !!! Warning
                All backend authentication used by Ravyn Simple JWT **must implement**
                the `async def authenticate()` functionality.
            """
        ),
    ]
    backend_refresh: Annotated[
        type[BaseRefreshAuthentication],
        Doc(
            """
            The backend refresh being used by the system. A subclass of `ravyn_simple_jwt.backends.BaseRefreshAuthentication`.

            !!! Warning
                All backend authentication used by Ravyn Simple JWT **must implement**
                the `async def refresh()` functionatility.
            """
        ),
    ]
    login_model: Annotated[
        type[BaseModel],
        Doc(
            """
            A pydantic base model with the needed fields for the login.
            Usually `email/username` and `password.

            This model can be found in `ravyn_simple_jwt.schemas.LoginEmailIn` and it is
            used by default for the login endpoint of a user into the system.

            !!! Tip
                If you don't want to use the default email/password but instead something
                unique to you, you can simply create your own model and override the `login_model`
                settings from the `SimpleJWT` configuration.
            """
        ),
    ] = LoginEmailIn
    refresh_model: Annotated[
        type[BaseModel],
        Doc(
            """
            A pydantic base model with the needed fields for the refresh token payload.
            Usually a dictionary of the format:

            ```python
            {
                "refresh_token": ...
            }
            ```

            This model can be found in `ravyn_simple_jwt.schemas.RefreshToken` and it is
            used by default for the refresh endpoint of an `access_token` in the system.
            """
        ),
    ] = RefreshToken
    access_token_model: Annotated[
        type[BaseModel],
        Doc(
            """
            **Used for OpenAPI specification and return of the refresh token URL**.

            A pydantic base model with the representing the return of an `access_token`:

            ```python
            {
                "access_token": ...
            }
            ```

            This model can be found in `ravyn_simple_jwt.schemas.AccessToken` and it is
            used by default for the refresh endpoint return of an `access_token` in the system.
            """
        ),
    ] = AccessToken
    token_model: Annotated[
        type[BaseModel],
        Doc(
            """
            **Used for OpenAPI specification only**.

            A pydantic base model with the representing the return of an `access_token` and `refresh_token`:

            ```python
            {
                "access_token": ...,
                "refresh_token": ...
            }
            ```

            This model can be found in `ravyn_simple_jwt.schemas.TokenAccess` and it is
            used by default for the refresh endpoint return of a dictionary containing the access and refresh tokens.
            """
        ),
    ] = TokenAccess
    tags: Annotated[
        str | None,
        Doc(
            """
            OpenAPI tags to be displayed on each view provided by Ravyn Simple JWT.

            These will be commmon to both views.
            """
        ),
    ] = None
    signin_url: Annotated[
        str,
        Doc(
            """
            The URL path in the format of `/path` used for the sign-in endpoint.
            """
        ),
    ] = "/signin"
    signin_summary: Annotated[
        str,
        Doc(
            """
            The OpenAPI URL summary for the path the sign-in endpoint.
            """
        ),
    ] = "Login API and returns a JWT Token."
    signin_description: Annotated[
        str,
        Doc(
            """
            The OpenAPI URL description for the path the sign-in endpoint.
            """
        ),
    ] = None
    refresh_url: Annotated[
        str,
        Doc(
            """
            The URL path in the format of `/path` used for the refresh token endpoint.
            """
        ),
    ] = "/refresh-access"
    refresh_summary: Annotated[
        str,
        Doc(
            """
            The OpenAPI URL summary for the path the refresh token endpoint.
            """
        ),
    ] = "Refreshes the access token"
    refresh_description: Annotated[
        str,
        Doc(
            """
            The OpenAPI URL description for the path the refresh token endpoint.
            """
        ),
    ] = """When a token expires, a new access token must be generated from the refresh token previously provided. The refresh token must be just that, a refresh and it should only return a new access token and nothing else
    """
    security: Annotated[
        list[Any] | None,
        Doc(
            """
            Used by OpenAPI definition, the security must be compliant with the norms.
            Ravyn offers some out of the box solutions where this is implemented.

            The [Ravyn security](https://ravyn.dev/openapi/) is available to automatically used.

            The security is applied to all the endpoints.
            ```
            """
        ),
    ] = [security]
